/*********************************************************************NVMH3****

Copyright NVIDIA Corporation 2002
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

******************************************************************************/

/*******************************************************************************
*                                                                              *
*   Basic example of the use of the CgFX runtime in a simple OpenGL program    *
*                                                                              *
*******************************************************************************/

/*******************************************************************************

    Headers

*******************************************************************************/

// C
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// CgFX
#include "CgFX/ICgFXEffect.h"

// OpenGL
#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glext.h>
#else
#include <GL/gl.h>
#include <GL/glext.h>
#endif

#ifdef WIN32
#define stricmp _stricmp
#endif

#ifdef linux
#define stricmp strcasecmp
#endif

// GLUT
#include "GL/glut.h"

// DDS
#include "nv_dds.h"

/*******************************************************************************

    Globals

*******************************************************************************/

/****** CgFX ******/

// Files
static char* EffectFile[] = { "Art/DiffuseSpecular.fx", 0 };
static unsigned int CurrentEffect;

// Interface
static ICgFXEffect* Effect;

// Techniques
#define EFFECT_TECHNIQUE_NUM_MAX 256
static bool TechniqueIsValid[EFFECT_TECHNIQUE_NUM_MAX];
static int TechniqueNum;
static int CurrentTechnique;

// Parameters
static CgFXPARAMETER_DESC ProjectionMatrixParam;
static CgFXPARAMETER_DESC ViewMatrixParam;
static CgFXPARAMETER_DESC ViewITMatrixParam;
static CgFXPARAMETER_DESC WorldMatrixParam;
static CgFXPARAMETER_DESC WorldITMatrixParam;
static CgFXPARAMETER_DESC WorldViewProjectionMatrixParam;
static CgFXPARAMETER_DESC WorldViewMatrixParam;
static CgFXPARAMETER_DESC WorldViewITMatrixParam;
static CgFXPARAMETER_DESC EyePositionParam;
static CgFXPARAMETER_DESC LightPositionParam;
static CgFXPARAMETER_DESC AmbientColorParam;
static CgFXPARAMETER_DESC DiffuseColorParam;
static CgFXPARAMETER_DESC SpecularColorParam;
static CgFXPARAMETER_DESC DiffuseMapParam;
static CgFXPARAMETER_DESC TweakableParam;

// Textures
#define EFFECT_TEXTURE_NUM_MAX 256
static unsigned int EffectTextureNum;
static GLuint EffectTexture[EFFECT_TEXTURE_NUM_MAX];


/****** Scene ******/

#define SCENE_OBJECT_NUM 2

// View parameters
static float ViewAngleX;
static float ViewAngleY;
static float ViewDistance = -10;

// Light position
static float LightPosition[] = { 50, 50, -50, 1 };

// World matrices
static float World[SCENE_OBJECT_NUM][4][4] = {
    {{1, 0, 0, 0},
     {0, 0,-1, 0},
     {0, 1, 0, 0},
     {0, 0, 0, 1}},
    {{1, 0, 0, 0},
     {0, 0, 1, 0},
     {0,-1, 0, 0},
     {0,-1, 0, 1}}};

// Colors
static float DiffuseColor[SCENE_OBJECT_NUM][3] = {{ 1, 1, 1 }, { 1, 1, 1 }};
static float SpecularColor[SCENE_OBJECT_NUM][3] = {{ 1, 1, 1 }, { 1, 1, 1 }};

// Textures
static const char* DiffuseMapName[SCENE_OBJECT_NUM] = {"Art/DiffuseMap0.dds", "Art/DiffuseMap1.dds"};
static GLuint DiffuseMap[SCENE_OBJECT_NUM];


/****** User interface ******/

static bool MouseLeftButtonDown;
static bool MouseRightButtonDown;
static int MouseLastX;
static int MouseLastY;
static float TweakableValue;
static float TweakableStep;
static float TweakableMin;
static float TweakableMax;

/*******************************************************************************

    Forward declarations

*******************************************************************************/

// CgFX
static bool SetupEffect(const char *effectFile);
static void SetParameters(unsigned int model, const float projection[4][4], const float view[4][4]);
static void Set4x4Matrix(CGFXHANDLE phandle, const float matrix[4][4]);
static void LoadEffectTextures(const CgFXEFFECT_DESC& effectDesc);
static void CheckForTweakableParameter(CGFXHANDLE paramID, CgFXPARAMETER_DESC* param);

// Scene
static void LoadSceneTextures();

// GLUT
static void Keyboard(unsigned char key, int x, int y);
static void Display();
static void Motion(int x, int y);
static void Mouse(int button, int state, int x, int y);

// Math
static void Transpose(const float A[4][4], float B[4][4]);
static float Det2x2(float a1, float a2, float b1, float b2);
static float Det3x3(float a1, float a2, float a3, float b1, float b2, float b3, float c1, float c2, float c3);
static void Invert(const float A[4][4], float B[4][4]);
static void Mult(const float A[4][4], const float B[4][4], float C[4][4]);
static void Mult(const float A[4][4], const float U[4], float V[4]);
static void Translate(float x, float y, float z, float T[4][4]);
static void RotateX(float a, float R[4][4]);
static void RotateY(float a, float R[4][4]);

/*******************************************************************************

    Entry point

*******************************************************************************/

int main(int argc, char *argv[])
{
    // Open window
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_ALPHA | GLUT_DEPTH);
    glutInitWindowSize(512, 512);
    glutCreateWindow("CgFX Simple");

    // Set callbacks
    glutKeyboardFunc(Keyboard);
    glutDisplayFunc(Display);
    glutMotionFunc(Motion);
    glutMouseFunc(Mouse);

    // Launch rendering loop
    glutMainLoop();

    return 0;
}

/*******************************************************************************

    Initialization

*******************************************************************************/

static void _cgfxError(FILE *fp, const char *str)
{
    const char *errstr = 0;
    fprintf(fp, str);
    if (FAILED(CgFXGetErrors(&errstr)))
        fprintf(fp, "unknown error\n");
    else
        fprintf(fp, "%s", errstr);
}

static bool SetupEffect(const char* effectFile)
{
    // Release previous effect
    if (Effect) {
        Effect->Release();
        Effect = 0;
    }
    ProjectionMatrixParam.Name = 0;
    ViewMatrixParam.Name = 0;
    WorldMatrixParam.Name = 0;
    WorldITMatrixParam.Name = 0;
    WorldViewProjectionMatrixParam.Name = 0;
    WorldViewMatrixParam.Name = 0;
    WorldViewITMatrixParam.Name = 0;
    ViewITMatrixParam.Name = 0;
    EyePositionParam.Name = 0;
    LightPositionParam.Name = 0;
    AmbientColorParam.Name = 0;
    DiffuseColorParam.Name = 0;
    SpecularColorParam.Name = 0;
    DiffuseMapParam.Name = 0;
    TweakableParam.Name = 0;

    // Load new effect
    const char* errors = 0;
    if (FAILED(CgFXCreateEffectFromFileA(effectFile, NULL, &Effect, &errors))) {
        printf("Cannot create effect: %s\n", errors);
        return false;
    }

    // Set OpenGL device
    if (FAILED(CgFXSetDevice("OpenGL", 0))) {
        _cgfxError(stderr, "Failed to set CgFX OpenGL device: ");
        return false;
    }

    // Get the effect description
    CgFXEFFECT_DESC effectDesc;
    if (FAILED(Effect->GetDesc(&effectDesc))) {
        _cgfxError(stderr, "Failed to get effect description: ");
        return false;
    }

    // Search for known parameters
    for (unsigned int i = 0; i < effectDesc.Parameters; ++i) {

        // Get parameter description
        CGFXHANDLE phandle = Effect->GetParameter(NULL, i);
        CgFXPARAMETER_DESC param;
        if (FAILED(Effect->GetParameterDesc(phandle, &param)))
            continue;

        // Search for a tweakable parameter
        CheckForTweakableParameter(phandle, &param);

        // Match parameter semantic to known semantic
        if (param.Semantic == 0) 
            continue;
        else if (stricmp(param.Semantic, "World") == 0) 
            WorldMatrixParam = param;
        else if (stricmp(param.Semantic, "WorldIT") == 0) 
            WorldITMatrixParam = param;
        else if (stricmp(param.Semantic, "WorldView") == 0) 
            WorldViewMatrixParam = param;
        else if (stricmp(param.Semantic, "WorldViewIT") == 0) 
            WorldViewITMatrixParam = param;
        else if (stricmp(param.Semantic, "WorldViewProjection") == 0) 
            WorldViewProjectionMatrixParam = param;
        else if (stricmp(param.Semantic, "Projection") == 0) 
            ProjectionMatrixParam = param;
        else if (stricmp(param.Semantic, "View") == 0) 
            ViewMatrixParam = param;
        else if (stricmp(param.Semantic, "ViewIT") == 0) 
            WorldViewITMatrixParam = param;
        else if (stricmp(param.Semantic, "EyePos") == 0) 
            EyePositionParam = param;
        else if (stricmp(param.Semantic, "LightPos") == 0) 
            LightPositionParam = param;
        else if (stricmp(param.Semantic, "Ambient") == 0) 
            AmbientColorParam = param;
        else if (stricmp(param.Semantic, "Diffuse") == 0) 
            DiffuseColorParam = param;
        else if (stricmp(param.Semantic, "Specular") == 0) 
            SpecularColorParam = param;
    }

    // Load the texture parameters
    LoadEffectTextures(effectDesc);

    // Search for valid techniques and chooses one
    TechniqueNum = effectDesc.Techniques;
    CurrentTechnique = -1;
    for (int t = 0; t < TechniqueNum; ++t) {

        // Set technique
        if (FAILED(Effect->SetTechnique(Effect->GetTechnique(t)))) {
            _cgfxError(stderr, "Failed to set technique: ");
            return false;
        }

        // Validate the technique
        TechniqueIsValid[t] = !FAILED(Effect->ValidateTechnique(Effect->GetCurrentTechnique()));
        if ((CurrentTechnique < 0) && TechniqueIsValid[t])
            CurrentTechnique = t;
    }

    // Leave if there is no valid technique
    if (CurrentTechnique < 0)
        return false;
    else {
        Effect->ValidateTechnique(Effect->GetCurrentTechnique());
    }

    return true;
}

static void LoadEffectTextures(const CgFXEFFECT_DESC& effectDesc)
{
    // Free previously loaded textures
    glDeleteTextures(EffectTextureNum, EffectTexture);
    EffectTextureNum = 0;

    // Load some default textures coming from the .fx file
    for (unsigned int i = 0; i < effectDesc.Parameters; ++i) {
        CgFXPARAMETER_DESC param;
        CGFXHANDLE paramHandle = Effect->GetParameter(NULL, i);
        HRESULT res = Effect->GetParameterDesc(paramHandle, &param);
        if (param.Type == CgFXPT_TEXTURE2D) {

            // Keep the diffuse map parameter, so we can set it to the diffuse texture provided with every model
            if (stricmp(param.Semantic, "DiffuseMap") == 0)
                DiffuseMapParam = param;

            // Load the texture associated to this parameter if a file is specified or...
            char* filename = 0;
            for (unsigned int j = 0; j < param.Annotations; ++j) {
                CGFXHANDLE ann = Effect->GetAnnotation(Effect->GetParameter(NULL, i), j);
                CgFXPARAMETER_DESC annotation;
                if (FAILED(Effect->GetParameterDesc(ann, &annotation)))
                    continue;

                if (strcmp(annotation.Name, "default") == 0 || 
                    strcmp(annotation.Name, "File") == 0) {
                    Effect->GetValue(ann, filename, annotation.Bytes);
                    break;
                }
            }
            // ... if the semantic is standard
            if (filename == 0 && stricmp(param.Semantic, "NormalizationMap") == 0)
                filename = "Art/NormalizationMap.dds";
            if (filename) {
                if (strcmp(".dds", filename + strlen(filename) - 4) != 0)
                    continue; // Not a DDS texture
                nv_dds::CDDSImage ddsimage;
                if (!ddsimage.load(filename))
                    continue; // Not loaded
                GLuint texture;
                glGenTextures(1, &texture);
                if (ddsimage.is_cubemap()) {
                    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, texture);
                    ddsimage.upload_textureCubemap();
                }
                else if (ddsimage.is_volume()) {
                    glBindTexture(GL_TEXTURE_3D, texture);
                    ddsimage.upload_texture3D();
                }
                else if (ddsimage.get_height() == 1)
                    ddsimage.upload_texture1D();
                else if (ddsimage.get_height() == ddsimage.get_width()) {
                    glBindTexture(GL_TEXTURE_2D, texture);
                    ddsimage.upload_texture2D();
                }
                else {
                    glBindTexture(GL_TEXTURE_RECTANGLE_NV, texture);
                    ddsimage.upload_textureRectangle();
                }
                Effect->SetTexture(paramHandle, (DWORD)texture);
                EffectTexture[EffectTextureNum] = texture;
                ++EffectTextureNum;
            }
        }
    }
}

static void LoadSceneTextures()
{
    // Free previously loaded textures
    glDeleteTextures(SCENE_OBJECT_NUM, DiffuseMap);

    // Load the scene diffuse textures
    for (unsigned int i = 0; i < SCENE_OBJECT_NUM; ++i) {
        if (strcmp(".dds", DiffuseMapName[i] + strlen(DiffuseMapName[i]) - 4) != 0)
            // Not a DDS texture
            continue;
        nv_dds::CDDSImage ddsimage;
        if (!ddsimage.load(DiffuseMapName[i]))
            // Not loaded
            continue;
        GLuint texture;
        glGenTextures(1, &texture);
        if (ddsimage.get_height() == ddsimage.get_width()) {
            glBindTexture(GL_TEXTURE_2D, texture);
            ddsimage.upload_texture2D();
        }
        else {
            glBindTexture(GL_TEXTURE_RECTANGLE_NV, texture);
            ddsimage.upload_textureRectangle();
        }
        DiffuseMap[i] = texture;
    }
}

// This functions illustrates how you detect parameters with a graphical user interface.
// For simplicity, we detect only the first one that matches some particular conditions.
static void CheckForTweakableParameter(CGFXHANDLE paramID, CgFXPARAMETER_DESC* param)
{
    if (TweakableParam.Name == 0 &&   // We haven't found one yet
        param->Type == CgFXPT_FLOAT && // It is a float parameter
        param->Rows == 1 && param->Columns == 1       // It is a vector or a scalar
        ) {

        // Examine the parameter annotation
        float maxFloat = 1e7;
        TweakableMin = - maxFloat;
        TweakableMax = maxFloat;
        TweakableStep = 0;
        for (unsigned int i = 0; i < param->Annotations; ++i) {
            CGFXHANDLE ann = Effect->GetAnnotation(paramID, i);
            CgFXPARAMETER_DESC annotation;
            if (FAILED(Effect->GetParameterDesc(ann, &annotation)))
                continue;
            if (annotation.Type == CgFXPT_FLOAT) {
                float value;
                Effect->GetValue(ann, &value, sizeof(float));

                if (stricmp(annotation.Name, "uimin") == 0)
                    TweakableMin = value;
                else if (stricmp(annotation.Name, "uimax") == 0)
                    TweakableMax = value;
                else if (stricmp(annotation.Name, "uistep") == 0)
                    TweakableStep = value;
            }
        }

        // Did we find one?
        if (TweakableMin != - maxFloat && TweakableMax != maxFloat && TweakableStep != 0) {
            TweakableParam = *param;
            float value[CGFX_MAX_DIMENSIONS];
            unsigned int size = max(1, (int)param->Elements);
            Effect->GetVector(paramID, value, size);
            TweakableValue = value[0]; // We will use only the first component to simplify
        }
    }
}

/*******************************************************************************

    Render 

*******************************************************************************/

// Function called every frame
static void Display() 
{
    static bool firstCall = true;

    // Set background color
    glClearColor(0.25f, 0.25f, 0.25f, 1.);

    if (firstCall) {
        // Initialize effect
        if (!SetupEffect(EffectFile[CurrentEffect])) {
            if (Effect)
                Effect->Release();
            exit(1);
        }
        firstCall = false;

        // Initialize scene
        LoadSceneTextures();
    }

    // Clear color and depth buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Compute the projection matrix
    float scale = 2.5f;
    float zNear = 0.1f;
    float zFar = 100;
    float projection[4][4] = {
        {scale,     0,                         0,  0},
        {    0, scale,                         0,  0},
        {    0,     0, (zFar+zNear)/(zNear-zFar), -1},
        {    0,     0, 2*zFar*zNear/(zNear-zFar),  0}};

    // Load the projection matrix
    if (ProjectionMatrixParam.Name)
        Set4x4Matrix(Effect->GetParameterByName(NULL, ProjectionMatrixParam.Name),
                     projection);

    // Compute the view matrix
    float view[4][4];
    float trans[4][4];
    Translate(0, 0, ViewDistance, trans);
    float rotx[4][4];
    RotateX(ViewAngleX, rotx);
    float roty[4][4];
    RotateY(ViewAngleY, roty);
    float tmp[4][4];
    Mult(trans, rotx, tmp);
    Mult(tmp, roty, view);

    // Load the view matrix
    if (ViewMatrixParam.Name)
        Set4x4Matrix(Effect->GetParameterByName(NULL, ViewMatrixParam.Name), view);

    // Load the inverse transpose of the view matrix
    if (ViewITMatrixParam.Name) {
        float viewI[4][4];
        Invert(view, viewI);
        float viewIT[4][4];
        Transpose(viewI, viewIT);
        Set4x4Matrix(Effect->GetParameterByName(NULL, ViewITMatrixParam.Name), viewIT);
    }

    // Render the scene with the chosen technique
    for (unsigned int m = 0; m < SCENE_OBJECT_NUM; ++m) {

        // Begin effect
        UINT numPasses;
        HRESULT hr = Effect->Begin(&numPasses, 0);

        // Set parameters
        SetParameters(m, projection, view);

        // Render the model with the effect
        for (unsigned int pass = 0; pass < numPasses; ++pass) {
            // Setup the pass
            hr = Effect->Pass(pass);

            // Sent the geometry
            glutSolidTeapot(1);
        }

        // End effect
        hr = Effect->End();
    }

    // Swap front and back buffers
    glutSwapBuffers();

    // Force display to be called again
    glutPostRedisplay();
}

// Set the rendering parameters for a particular model of the scene
static void SetParameters(unsigned int model, const float projection[4][4], const float view[4][4])
{
    // Set the world matrix
    if (WorldMatrixParam.Name)
        Set4x4Matrix(Effect->GetParameterByName(NULL, WorldMatrixParam.Name), World[model]);

    // Compute and set the inverse transpose of the world matrix
    if (WorldITMatrixParam.Name) {
        float worldI[4][4];
        Invert(World[model], worldI);
        float worldIT[4][4];
        Transpose(worldI, worldIT);
        Set4x4Matrix(Effect->GetParameterByName(NULL, WorldITMatrixParam.Name), worldIT);
    }

    // Compute and set the world view matrix
    float worldView[4][4];
    Mult(view, World[model], worldView);
    if (WorldViewMatrixParam.Name)
        Set4x4Matrix(Effect->GetParameterByName(NULL, WorldViewMatrixParam.Name), worldView);

    // Compute and set the inverse transpose of the world view matrix
    if (WorldViewITMatrixParam.Name || EyePositionParam.Name) {
        float worldViewI[4][4];
        Invert(worldView, worldViewI);
        if (EyePositionParam.Name)
            Effect->SetVector(Effect->GetParameterByName(NULL, EyePositionParam.Name),
                              &worldViewI[3][0], 3);
        float worldViewIT[4][4];
        Transpose(worldViewI, worldViewIT);
        if (WorldViewITMatrixParam.Name)
            Set4x4Matrix(Effect->GetParameterByName(NULL, WorldViewITMatrixParam.Name),
                         worldViewIT);
    }

    // Compute and set the world view projection matrix
    if (WorldViewProjectionMatrixParam.Name) {
        float worldViewProjection[4][4];
        Mult(projection, worldView, worldViewProjection);
        Set4x4Matrix(Effect->GetParameterByName(NULL, WorldViewProjectionMatrixParam.Name),
                     worldViewProjection);
    }

    // Set light position
    if (LightPositionParam.Name) {
        float lightPosInModelSpace[4];
        Mult(World[model], LightPosition, lightPosInModelSpace);
        Effect->SetVector(Effect->GetParameterByName(NULL, LightPositionParam.Name),
                          lightPosInModelSpace, 3);
    }

    // Set colors
    if (DiffuseColorParam.Name)
        Effect->SetVector(Effect->GetParameterByName(NULL, DiffuseColorParam.Name),
                          DiffuseColor[model], 3);
    if (SpecularColorParam.Name)
        Effect->SetVector(Effect->GetParameterByName(NULL, SpecularColorParam.Name),
                          SpecularColor[model], 3);

    // Set textures
    if (DiffuseMapParam.Name)
        Effect->SetTexture(Effect->GetParameterByName(NULL, DiffuseMapParam.Name),
                           (DWORD)DiffuseMap[model]);

    // Set the tweakable parameter
    if (TweakableParam.Name) {
        float value[CGFX_MAX_DIMENSIONS];
        for (unsigned int i = 0; i < CGFX_MAX_DIMENSIONS; ++i)
            value[i] = TweakableValue;
        unsigned int size = max(1, (int)TweakableParam.Elements);
        Effect->SetVector(Effect->GetParameterByName(NULL, TweakableParam.Name),
                          value, size);
    }
}

static void Set4x4Matrix(CGFXHANDLE phandle, const float matrix[4][4])
{
    // CgFX assumes row-major matrices and we're manipulating OpenGL column-major matrices
    float matrixT[4][4];
    Transpose(matrix, matrixT);
    Effect->SetMatrix(phandle, &matrixT[0][0], 4, 4);
}

/*******************************************************************************

    I/O 

*******************************************************************************/

static void Keyboard(unsigned char key, int x, int y)
{
    // Quit application
    if (key == 'q' || key == 'Q' || key == 27 /* escape */)
        exit(0);

    // Change the value of the tweakable parameter
    else if (key == '+') {
        TweakableValue += TweakableStep;
        if (TweakableValue < TweakableMin)
            TweakableValue = TweakableMin;
        else if (TweakableMax < TweakableValue)
            TweakableValue = TweakableMax;
    }
    else if (key == '-') {
        TweakableValue -= TweakableStep;
        if (TweakableValue < TweakableMin)
            TweakableValue = TweakableMin;
        else if (TweakableMax < TweakableValue)
            TweakableValue = TweakableMax;
    }

    // Change the technique used
    else if (key == ' ') {
        do {
            ++CurrentTechnique;
            if (CurrentTechnique == TechniqueNum)
                CurrentTechnique = 0;
        } while (!TechniqueIsValid[CurrentTechnique]);
        Effect->SetTechnique(Effect->GetTechnique(CurrentTechnique));
        Effect->ValidateTechnique(Effect->GetCurrentTechnique());
    }
}

static void Motion(int x, int y)
{
    // Rotate the view
    if (MouseLeftButtonDown) {
        float speed = 0.02f;
        ViewAngleY += speed * (x - MouseLastX);
        ViewAngleX += speed * (y - MouseLastY);
    }

    // Zoom in and out
    else if (MouseRightButtonDown) {
        float speed = 0.6f;
        ViewDistance += speed * (y - MouseLastY);
    }

    MouseLastX = x;
    MouseLastY = y;
}

static void Mouse(int button, int state, int x, int y)
{
    bool buttonDown = MouseLeftButtonDown || MouseRightButtonDown ? false : state == GLUT_DOWN;
    switch (button) {
    case GLUT_LEFT_BUTTON:
        MouseLeftButtonDown = buttonDown;
        break;
    case GLUT_RIGHT_BUTTON:
        MouseRightButtonDown = buttonDown;
        break;
    default:
        break;
    }
    if (buttonDown) {
        MouseLastX = x;
        MouseLastY = y;
    }
}

/*******************************************************************************

    Math

*******************************************************************************/

// Matrices are assumed to be stored in column major order like OpenGL does

static void Transpose(const float A[4][4], float B[4][4])
{
    for (unsigned int i = 0; i < 4; ++i)
        for (unsigned int j = 0; j < 4; ++j)
            B[i][j] = A[j][i];
}

static float Det2x2(float a1, float a2, float b1, float b2)
{
    return a1 * b2 - b1 * a2;
}

static float Det3x3(float a1, float a2, float a3, float b1, float b2, float b3, float c1, float c2, float c3)
{
    return a1 * Det2x2(b2, b3, c2, c3) - b1 * Det2x2(a2, a3, c2, c3) + c1 * Det2x2(a2, a3, b2, b3);
}

static void Invert(const float A[4][4], float B[4][4])
{
    B[0][0] =  Det3x3(A[1][1], A[1][2], A[1][3], A[2][1], A[2][2], A[2][3], A[3][1], A[3][2], A[3][3]);
    B[0][1] = -Det3x3(A[0][1], A[0][2], A[0][3], A[2][1], A[2][2], A[2][3], A[3][1], A[3][2], A[3][3]);
    B[0][2] =  Det3x3(A[0][1], A[0][2], A[0][3], A[1][1], A[1][2], A[1][3], A[3][1], A[3][2], A[3][3]);
    B[0][3] = -Det3x3(A[0][1], A[0][2], A[0][3], A[1][1], A[1][2], A[1][3], A[2][1], A[2][2], A[2][3]);
    B[1][0] = -Det3x3(A[1][0], A[1][2], A[1][3], A[2][0], A[2][2], A[2][3], A[3][0], A[3][2], A[3][3]);
    B[1][1] =  Det3x3(A[0][0], A[0][2], A[0][3], A[2][0], A[2][2], A[2][3], A[3][0], A[3][2], A[3][3]);
    B[1][2] = -Det3x3(A[0][0], A[0][2], A[0][3], A[1][0], A[1][2], A[1][3], A[3][0], A[3][2], A[3][3]);
    B[1][3] =  Det3x3(A[0][0], A[0][2], A[0][3], A[1][0], A[1][2], A[1][3], A[2][0], A[2][2], A[2][3]);
    B[2][0] =  Det3x3(A[1][0], A[1][1], A[1][3], A[2][0], A[2][1], A[2][3], A[3][0], A[3][1], A[3][3]);
    B[2][1] = -Det3x3(A[0][0], A[0][1], A[0][3], A[2][0], A[2][1], A[2][3], A[3][0], A[3][1], A[3][3]);
    B[2][2] =  Det3x3(A[0][0], A[0][1], A[0][3], A[1][0], A[1][1], A[1][3], A[3][0], A[3][1], A[3][3]);
    B[2][3] = -Det3x3(A[0][0], A[0][1], A[0][3], A[1][0], A[1][1], A[1][3], A[2][0], A[2][1], A[2][3]);
    B[3][0] = -Det3x3(A[1][0], A[1][1], A[1][2], A[2][0], A[2][1], A[2][2], A[3][0], A[3][1], A[3][2]);
    B[3][1] =  Det3x3(A[0][0], A[0][1], A[0][2], A[2][0], A[2][1], A[2][2], A[3][0], A[3][1], A[3][2]);
    B[3][2] = -Det3x3(A[0][0], A[0][1], A[0][2], A[1][0], A[1][1], A[1][2], A[3][0], A[3][1], A[3][2]);
    B[3][3] =  Det3x3(A[0][0], A[0][1], A[0][2], A[1][0], A[1][1], A[1][2], A[2][0], A[2][1], A[2][2]);
    float det = (A[0][0] * B[0][0]) + (A[1][0] * B[0][1]) + (A[2][0] * B[0][2]) + (A[3][0] * B[0][3]);
    det = 1 / det;
    B[0][0] *= det;
    B[0][1] *= det;
    B[0][2] *= det;
    B[0][3] *= det;
    B[1][0] *= det;
    B[1][1] *= det;
    B[1][2] *= det;
    B[1][3] *= det;
    B[2][0] *= det;
    B[2][1] *= det;
    B[2][2] *= det;
    B[2][3] *= det;
    B[3][0] *= det;
    B[3][1] *= det;
    B[3][2] *= det;
    B[3][3] *= det;
}

static void Mult(const float A[4][4], const float B[4][4], float C[4][4])
{
    C[0][0] = A[0][0] * B[0][0] + A[1][0] * B[0][1] + A[2][0] * B[0][2] + A[3][0] * B[0][3];
    C[0][1] = A[0][1] * B[0][0] + A[1][1] * B[0][1] + A[2][1] * B[0][2] + A[3][1] * B[0][3];
    C[0][2] = A[0][2] * B[0][0] + A[1][2] * B[0][1] + A[2][2] * B[0][2] + A[3][2] * B[0][3];
    C[0][3] = A[0][3] * B[0][0] + A[1][3] * B[0][1] + A[2][3] * B[0][2] + A[3][3] * B[0][3];
    C[1][0] = A[0][0] * B[1][0] + A[1][0] * B[1][1] + A[2][0] * B[1][2] + A[3][0] * B[1][3];
    C[1][1] = A[0][1] * B[1][0] + A[1][1] * B[1][1] + A[2][1] * B[1][2] + A[3][1] * B[1][3];
    C[1][2] = A[0][2] * B[1][0] + A[1][2] * B[1][1] + A[2][2] * B[1][2] + A[3][2] * B[1][3];
    C[1][3] = A[0][3] * B[1][0] + A[1][3] * B[1][1] + A[2][3] * B[1][2] + A[3][3] * B[1][3];
    C[2][0] = A[0][0] * B[2][0] + A[1][0] * B[2][1] + A[2][0] * B[2][2] + A[3][0] * B[2][3];
    C[2][1] = A[0][1] * B[2][0] + A[1][1] * B[2][1] + A[2][1] * B[2][2] + A[3][1] * B[2][3];
    C[2][2] = A[0][2] * B[2][0] + A[1][2] * B[2][1] + A[2][2] * B[2][2] + A[3][2] * B[2][3];
    C[2][3] = A[0][3] * B[2][0] + A[1][3] * B[2][1] + A[2][3] * B[2][2] + A[3][3] * B[2][3];
    C[3][0] = A[0][0] * B[3][0] + A[1][0] * B[3][1] + A[2][0] * B[3][2] + A[3][0] * B[3][3];
    C[3][1] = A[0][1] * B[3][0] + A[1][1] * B[3][1] + A[2][1] * B[3][2] + A[3][1] * B[3][3];
    C[3][2] = A[0][2] * B[3][0] + A[1][2] * B[3][1] + A[2][2] * B[3][2] + A[3][2] * B[3][3];
    C[3][3] = A[0][3] * B[3][0] + A[1][3] * B[3][1] + A[2][3] * B[3][2] + A[3][3] * B[3][3];
}

static void Mult(const float A[4][4], const float U[4], float V[4])
{
    V[0] = A[0][0] * U[0] + A[1][0] * U[1] + A[2][0] * U[2] + A[3][0] * U[3];
    V[1] = A[0][1] * U[0] + A[1][1] * U[1] + A[2][1] * U[2] + A[3][1] * U[3];
    V[2] = A[0][2] * U[0] + A[1][2] * U[1] + A[2][2] * U[2] + A[3][2] * U[3];
    V[3] = A[0][3] * U[0] + A[1][3] * U[1] + A[2][3] * U[2] + A[3][3] * U[3];
}

static void Translate(float x, float y, float z, float T[4][4])
{
    T[0][0] = 1; T[0][1] = 0; T[0][2] = 0; T[0][3] = 0;
    T[1][0] = 0; T[1][1] = 1; T[1][2] = 0; T[1][3] = 0;
    T[2][0] = 0; T[2][1] = 0; T[2][2] = 1; T[2][3] = 0;
    T[3][0] = x; T[3][1] = y; T[3][2] = z; T[3][3] = 1;
}

static void RotateX(float a, float R[4][4])
{
    float c = cosf(a);
    float s = sinf(a);
    R[0][0] = 1; R[0][1] = 0;  R[0][2] = 0; R[0][3] = 0;
    R[1][0] = 0; R[1][1] = c;  R[1][2] = s; R[1][3] = 0;
    R[2][0] = 0; R[2][1] = -s; R[2][2] = c; R[2][3] = 0;
    R[3][0] = 0; R[3][1] = 0;  R[3][2] = 0; R[3][3] = 1;
}

static void RotateY(float a, float R[4][4])
{
    float c = cosf(a);
    float s = sinf(a);
    R[0][0] = c; R[0][1] = 0; R[0][2] = -s; R[0][3] = 0;
    R[1][0] = 0; R[1][1] = 1; R[1][2] = 0;  R[1][3] = 0;
    R[2][0] = s; R[2][1] = 0; R[2][2] = c;  R[2][3] = 0;
    R[3][0] = 0; R[3][1] = 0; R[3][2] = 0;  R[3][3] = 1;
}
